/*
  ======================================================================
   demo.c --- protoype to show off the simple solver
  ----------------------------------------------------------------------
   Author : Jos Stam (jstam@aw.sgi.com)
   Creation Date : Jan 9 2003

   Description:

	This code is a simple prototype that demonstrates how to use the
	code provided in my GDC2003 paper entitles "Real-Time Fluid Dynamics
	for Games". This code uses OpenGL and GLUT for graphics and interface

  =======================================================================
*/

#include <stdlib.h>
#include <stdio.h>
#include <nehe.h>
#include "stam.h"
#include <sven.h>

namespace display {
bool show_bounds=true;
bool show_velocity=false;
bool show_density=true;

//npr-ish stuff
bool show_points=true;
bool show_brushes=false;
bool show_triangles=false;
}

namespace Stam {

using namespace display;
using sven::float3;
using sven::rgb;

rgb bound_color(255,255,255);

/* macros */

#define IX(i,j) ((i)+(N+2)*(j))
#define FOR_EACH_CELL for ( i=1 ; i<=N ; i++ ) { for ( j=1 ; j<=N ; j++ ) { 
#define END_FOR }}

/* external definitions (from solver.c) */

extern void dens_step ( int N, float * x, float * x0, float * u, float * v, float diff, float dt );
extern void vel_step ( int N, float * u, float * v, float * u0, float * v0, float visc, float dt );

// global variables

int N;
float dt, diff, visc;
float ddecay, dalpha;
float force, source;  // should probably add these to the UI as well...


float * u, * v, * u_prev, * v_prev;


const int num_dyes = 4;
int selected_dye=0;
float * dyes[num_dyes*2];
float3 dcolor[num_dyes] = {float3(1,1,1), float3(1,0,0), float3(0,0,1), float3(0,1,0)};
//float3 dcolor[num_dyes] = {float3(1,0,0)};

static int win_id;
int win_x, win_y;
static int mouse_down[3];
static int omx, omy, mx, my;



//  ----------------------------------------------------------------------
//   free/clear/allocate simulation data
//  ----------------------------------------------------------------------


void free_data ( void )
{
	if ( u ) free ( u );
	if ( v ) free ( v );
	if ( u_prev ) free ( u_prev );
	if ( v_prev ) free ( v_prev );

	for(int i=0;i<2*num_dyes;i++) {
		if(dyes[i]) free(dyes[i]);
	}

	if (blocked) free (blocked);
}

void clear_data ( void )
{
	int i, size=(N+2)*(N+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = u_prev[i] = v_prev[i] = 0.0f;
		for(int j=0;j<2*num_dyes;j++) {
			dyes[j][i] =  0.0f;
		}
		blocked[i]=false;
	}
}



static int allocate_data ( void )
{
	int size = (N+2)*(N+2);

	blocked		= (bool *)  malloc ( size*sizeof(bool) );
	u			= (float *) malloc ( size*sizeof(float) );
	v			= (float *) malloc ( size*sizeof(float) );
	u_prev		= (float *) malloc ( size*sizeof(float) );
	v_prev		= (float *) malloc ( size*sizeof(float) );
	
	
	for(int i=0;i<2*num_dyes;i++) {
		dyes[i] =(float *) malloc ( size*sizeof(float) );
	}
	if ( !u || !v || !u_prev || !v_prev || !blocked) {
		fprintf ( stderr, "cannot allocate data\n" );
		return ( 0 );
	}

	return ( 1 );
}

void init_data() {
	N = 64;
	dt = 0.1f;
	diff = 0.0f;
	visc = 0.000f;
	
	ddecay=1.0f;
	dalpha=3.0f;
	
	force = 15.0f;
	source = 100.0f;

	if ( !allocate_data () ) exit ( 1 );
	clear_data ();
}


//  ----------------------------------------------------------------------
//   OpenGL specific drawing routines
//  ----------------------------------------------------------------------


static void pre_display ( void )
{
	glViewport ( 0, 0, win_x, win_y );
	glMatrixMode ( GL_PROJECTION );
	glLoadIdentity ();
	gluOrtho2D ( 0.0, 1.0, 0.0, 1.0 );
	glClearColor ( 0.0f, 0.0f, 0.0f, 1.0f );
}


static void draw_velocity ( void )
{
	int i, j;
	float x, y, h;

	h = 1.0f/N;

	glColor3f ( 1.0f, 1.0f, 1.0f );
	glLineWidth ( 1.0f );

	glBegin ( GL_LINES );

		for ( i=1 ; i<=N ; i++ ) {
			x = (i-0.5f)*h;
			for ( j=1 ; j<=N ; j++ ) {
				y = (j-0.5f)*h;

				glVertex2f ( x, y );
				glVertex2f ( x+u[IX(i,j)], y+v[IX(i,j)] );
			}
		}

	glEnd ();
}

void draw_bounds () {
	if(!show_bounds) return;
	int i,j;
	glBegin(GL_QUADS);
	float x,y,h;
	h = 1.0f/N;
	for ( i=0 ; i<=N ; i++ ) {
		x = (i-0.5f)*h;
		for ( j=0 ; j<=N ; j++ ) {
		y = (j-0.5f)*h;	
			if(blocked[IX(i,j)]) {	
				glColor4f ( bound_color.r/255.,bound_color.g/255.,bound_color.b/255.,1. ); 
				glVertex2f ( x, y);
				glVertex2f ( x+h, y );
				glVertex2f ( x+h, y+h );
				glVertex2f ( x,  y+h);		
			}
		}
	}
	glEnd();
}

void draw_density ( void )
{
	int i, j;
	float x, y, h, d00, d01, d10, d11;

	h = 1.0f/N;

	glBegin ( GL_QUADS );
	for( int k=0;k<num_dyes;k++){
		for ( i=0 ; i<=N ; i++ ) {
			x = (i-0.5f)*h;
			for ( j=0 ; j<=N ; j++ ) {
				y = (j-0.5f)*h;

				d00 = dyes[2*k][IX(i,j)]     * dalpha;
				d01 = dyes[2*k][IX(i,j+1)]   * dalpha ;
				d10 = dyes[2*k][IX(i+1,j)]   * dalpha;
				d11 = dyes[2*k][IX(i+1,j+1)] * dalpha;

				glColor4f ( d00*dcolor[k][0], d00*dcolor[k][1], d00*dcolor[k][2], d00 ); glVertex2f ( x, y );
				glColor4f ( d10*dcolor[k][0], d10*dcolor[k][1], d10*dcolor[k][2], d10 ); glVertex2f ( x+h, y );
				glColor4f ( d11*dcolor[k][0], d11*dcolor[k][1], d11*dcolor[k][2], d11 ); glVertex2f ( x+h, y+h );
				glColor4f ( d01*dcolor[k][0], d01*dcolor[k][1], d01*dcolor[k][2], d01 ); glVertex2f ( x, y+h );
			}
		}

		if(&boundsUI==currentUI) {
			int i = (int)((       mx /(float)win_x)*N+1);
			int j = (int)(((win_y-my)/(float)win_y)*N+1);

			//if((i%2)) i++;
			//if((j%2)) j++;

			if ( !(i<2 || i>N || j<2 || j>N) ) {

			x = (i-0.5f)*h;
			y = (j-0.5f)*h;

			glColor4f ( bound_color.r,bound_color.g,bound_color.b,0.2f ); 
			glVertex2f ( x-h, y-h );
			glVertex2f ( x+h, y-h );
			glVertex2f ( x+h, y+h );
			glVertex2f ( x-h, y+h );
			}
		}
	}
	glEnd ();
}

void set_bound( int mx, int my) {
	int i = (int)((       mx /(float)win_x)*N+1);
	int j = (int)(((win_y-my)/(float)win_y)*N+1);

	//if((i%2)) i++;
	//if((j%2)) j++;

	if ( i<2 || i>N || j<2 || j>N ) return;

	blocked[IX(i,j)]=true;
	blocked[IX(i-1,j)]=true;
	blocked[IX(i,j-1)]=true;
	blocked[IX(i-1,j-1)]=true;
}


void kill_bound( int mx, int my) {
	int i = (int)((       mx /(float)win_x)*N+1);
	int j = (int)(((win_y-my)/(float)win_y)*N+1);

	//if((i%2)) i++;
	//if((j%2)) j++;

	if ( i<2 || i>N || j<2 || j>N ) return;

	blocked[IX(i,j)]=false;
	blocked[IX(i-1,j)]=false;
	blocked[IX(i,j-1)]=false;
	blocked[IX(i-1,j-1)]=false;
}

//  ----------------------------------------------------------------------
//   relates mouse movements to forces sources
//  ----------------------------------------------------------------------


static void get_from_UI ( float * d, float * u, float * v )
{
	int i, j, size = (N+2)*(N+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = d[i] = 0.0f;
	}

	if ( !mouse_down[0] && !mouse_down[2] ) goto end;

	i = (int)((       mx /(float)win_x)*N+1);
	j = (int)(((win_y-my)/(float)win_y)*N+1);

	if ( i<1 || i>N || j<1 || j>N ) goto end;

	if ( mouse_down[0] ) {
		u[IX(i,j)] = force * (mx-omx);
		v[IX(i,j)] = force * (omy-my);
	}

	if ( mouse_down[2] ) {
		d[IX(i,j)] = source;
	}

end:
	omx = mx;
	omy = my;

	return;
}

/*
  ----------------------------------------------------------------------
   GLUT callback routines
  ----------------------------------------------------------------------
*/

void key_func ( unsigned char key, int x, int y )
{
	if(key-'0' > 0 && key-'0' <= num_dyes ) {
		selected_dye=key-'1';
		//ostringstream i;
		//i<<selected_dye;
		//win_tools::fyi_msg(i.str().c_str());
	}
	
	switch ( key )
	{
		case 'c':
		case 'C':
			clear_data ();
			break;


		case 'b':
		case 'B':
			show_bounds=!show_bounds;
			break;


		case 'v':
		case 'V':
			//dvel = !dvel;
			break;
	}
}

void mouse_func ( int button, bool down, int x, int y )
{
	 mx = x;
	 my = y;

	mouse_down[button] = down;
}

void motion_func ( int x, int y )
{
	mx = x;
	my = y;
}

void reshape_func ( int width, int height )
{
	win_x = width;
	win_y = height;
}


void add_vorocity(int N, float * u, float * v, float dt);

void idle_func ( void )
{

	//this density decay trick can make for nicer visualizations
	int i,j;
	FOR_EACH_CELL 
		for(int k=0;k<num_dyes;k++) {
			dyes[2*k][IX(i,j)]*=ddecay;
			dyes[2*k+1][IX(i,j)] = 0.0f;
		}
	END_FOR

	using namespace display;

	get_from_UI ( dyes[2*selected_dye+1], u_prev, v_prev );
	add_vorocity(N, u, v, dt); 
	vel_step ( N, u, v, u_prev, v_prev, visc, dt );

	
	for(i=0;i<num_dyes;i++) {
		dens_step ( N, dyes[2*i], dyes[2*i+1], u, v, diff, dt );
	}
}

void display_func ( void )
{
	pre_display ();
	if ( show_density )	draw_density ();
	if ( show_velocity) draw_velocity ();
}

}